/* Model of a forest fire - a 2D rectangular grid of trees is initialized
   with one random tree caught on fire. At each time step, trees that are not
   on fire yet check their neighbors to the north, east, south, and west, and
   if any of them are on fire, the tree catches fire with some percent chance.
   The model runs for a certain number of time steps, which can be controlled
   by the user. At the end of the simulation, the program outputs the total
   percentage of trees burned. Tree data can also be output at each time step
   if a filename is provided by the user.

   All processes are responsible for running the
   simulation and outputting data. The rows of the forest are split up as
   evenly as possible among the processes, with the remainder split evenly among
   the first few processes. At each time step, processes must each send one row 
   of data to their top neighbor and one row of data to their bottom neighbor to
   communicate at the boundaries between process' rows. Code that should only be
   executed by one process are performed by a single "boss" process.
   */

/* Author: Aaron Weeden, Shodor, 2015 */

/* Naming convention:
   ALL_CAPS for constants
   CamelCase for globals and functions
   lowerCase for locals
   */
#include <mpi.h> /* MPI_Init(), MPI_Finalize(), etc. */
#include <stdbool.h> /* bool type */
#include <stdio.h> /* printf() */
#include <stdlib.h> /* atoi(), exit(), EXIT_FAILURE, malloc(), free(),
                       random() */
#include <string.h> /* strcpy() */
#include <unistd.h> /* getopt() */

/* Declare which MPI process is the boss */
#define BOSS 0

/* Define descriptions of command line options */
#define ROWS_DESCR \
  "The forest has this many rows of trees (positive integer)"
#define COLS_DESCR \
  "The forest has this many columns of trees (positive integer)"
#define BURN_PROB_DESCR \
  "Chance of catching fire if next to burning tree (positive integer [0..100])"
#define MAX_BURN_STEPS_DESCR \
  "A burning tree stops burning after this many time steps (positive integer)"
#define STEPS_DESCR \
  "Run for this many time steps (positive integer)"
#define RAND_SEED_DESCR \
  "Seed value for the random number generator (positive integer)"
#define OUTPUT_FILENAME_DESCR \
  "Filename to output tree data at each time step (file must not already exist)"

/* Define default values for simulation parameters - each of these parameters
   can also be changed later via user input */
#define DEFAULT_ROWS 21
#define DEFAULT_COLS 21
#define DEFAULT_BURN_PROB 100
#define DEFAULT_MAX_BURN_STEPS 5
#define DEFAULT_STEPS 10
#define DEFAULT_RAND_SEED 1
#define DEFAULT_OUTPUT_FILENAME "fire-out.txt"
#define DEFAULT_IS_OUTPUTTING_EACH_STEP false

/* Define characters used on the command line to change the values of input
   parameters */
#define ROWS_CHAR 'r'
#define COLS_CHAR 'c'
#define BURN_PROB_CHAR 'b'
#define MAX_BURN_STEPS_CHAR 'm'
#define STEPS_CHAR 't'
#define RAND_SEED_CHAR 's'
#define OUTPUT_FILENAME_CHAR 'o'

/* Define options string used by getopt() - a colon after the character means
   the parameter's value is specified by the user */
const char GETOPT_STRING[] = {
  ROWS_CHAR, ':',
  COLS_CHAR, ':',
  BURN_PROB_CHAR, ':',
  MAX_BURN_STEPS_CHAR, ':',
  STEPS_CHAR, ':',
  RAND_SEED_CHAR, ':',
  OUTPUT_FILENAME_CHAR, ':'
};

/* Define a mapping from the row and column of a given tree in a forest with
   boundaries to the index of that tree in a 1D array that includes
   boundaries */
#define TREE_MAP(row, col, ncols) ((row) * (ncols + 2) + (col))

/* Define a mapping from the row and column of a given tree in a forest with
   boundaries to the index of that tree in a 1D array that does not include
   boundaries */
#define NEW_TREE_MAP(row, col, ncols) ((row - 1) * (ncols) + (col - 1))

/* Declare global parameters */
int Rows = DEFAULT_ROWS;
int Cols = DEFAULT_COLS;
int BurnProb = DEFAULT_BURN_PROB;
int MaxBurnSteps = DEFAULT_MAX_BURN_STEPS;
int Steps = DEFAULT_STEPS;
int RandSeed = DEFAULT_RAND_SEED;
bool IsOutputtingEachStep = DEFAULT_IS_OUTPUTTING_EACH_STEP;
char *OutputFilename = DEFAULT_OUTPUT_FILENAME;

/* Declare other needed global variables */
bool AreParamsValid = true; /* Do the model parameters have valid values? */
int CurStep; /* The current time step */
int BurnedTrees = 0; /* The total number of burned trees */
char ExeName[32]; /* The name of the program executable */
int MaxBurnStepsDigits; /* The number of digits in the max burn steps; used for
                           outputting tree data */
int *Trees; /* 1D tree array, contains a boundary around the outside of the
               forest so the same neighbor checking algorithm can be used on
               all cells */
int *NewTrees; /* Copy of 1D tree array - used so that we don't update the
                  forest too soon as we are deciding which new trees
                  should burn -- does not contain boundary */
int MyRank; /* MPI rank of the given process executing this code */
int Procs; /* Total number of MPI processes */
int *SendBuf; /* Buffer for sending messages using MPI */
int *RecvBuf; /* Buffer for receiving messages using MPI */
FILE *OutputFile; /* For outputting tree data to a file */

/* DECLARE FUNCTIONS */

/* Prints out a description of an integer command line option

   @param optChar The character used to specify the option
   @param optDescr The description of the option
   @param optDefault The default value of the option
   */
void DescribeOptionInt(const char optChar, const char *optDescr,
    const int optDefault) {
  fprintf(stderr, "-%c : \n\t%s\n\tdefault: %d\n", optChar, optDescr,
      optDefault);
}

/* Prints out a description of a string command line option

   @param optChar The character used to specify the option
   @param optDescr The description of the option
   @param optDefault The default value of the option
   */
void DescribeOptionStr(const char optChar, const char *optDescr,
    const char *optDefault) {
  fprintf(stderr, "-%c : \n\t%s\n\tdefault: %s\n", optChar, optDescr,
      optDefault);
}

/* Print an error message

   @param errorMsg Buffer containing the message
   */
void PrintError(const char *errorMsg) {
  if (MyRank == BOSS) {
    fprintf(stderr, "%s", errorMsg);
  }
}

/* Display to the user what options are available for running the program and
   exit the program in failure 

   @param errorMsg The error message to print
   */
void PrintUsageAndExit() {
  if (MyRank == BOSS) {
    fprintf(stderr, "Usage: ");
    fprintf(stderr, "%s [OPTIONS]\n", ExeName);
    fprintf(stderr, "Where OPTIONS can be any of the following:\n");
    DescribeOptionInt(ROWS_CHAR, ROWS_DESCR, DEFAULT_ROWS);
    DescribeOptionInt(COLS_CHAR, COLS_DESCR, DEFAULT_COLS);
    DescribeOptionInt(BURN_PROB_CHAR, BURN_PROB_DESCR, DEFAULT_BURN_PROB);
    DescribeOptionInt(MAX_BURN_STEPS_CHAR, MAX_BURN_STEPS_DESCR,
        DEFAULT_MAX_BURN_STEPS);
    DescribeOptionInt(STEPS_CHAR, STEPS_DESCR,
        DEFAULT_STEPS);
    DescribeOptionInt(RAND_SEED_CHAR, RAND_SEED_DESCR, DEFAULT_RAND_SEED);
    DescribeOptionStr(OUTPUT_FILENAME_CHAR, OUTPUT_FILENAME_DESCR,
        DEFAULT_OUTPUT_FILENAME);
  }
  MPI_Finalize();
  exit(EXIT_FAILURE);
}

/* Assert that a user's input value is a positive integer. If it is not, print
   a usage message and an error message and exit the program.

   @param param The user's input value
   @param optChar The character used the specify the user's input value
   */
void AssertPositiveInteger(int param, const char optChar) {
  char errorStr[64];

  /* Get the user's input value, assume floating point */
  const float floatParam = atof(optarg);

  /* Make sure positive and integer */
  if (param < 1 || floatParam != param) {
    sprintf(errorStr, "ERROR: value for -%c must be positive integer\n",
        optChar);
    PrintError(errorStr);
    AreParamsValid = false;
  }
}

/* Assert that a user's input value is between two values, inclusive. If it is
   not, print a usage message and an error message and exit the program.

   @param param The user's input value
   @param low The lowest value the parameter can be
   @param high The highest value the parameter can be
   @param optChar The character used the specify the user's input value
   */
void AssertBetweenInclusive(int param, const int low, const int high,
    const char optChar) {
  char errorStr[64];

  if (param < low || param > high) {
    sprintf(errorStr,
        "ERROR: value for -%c must be between %d and %d, inclusive\n",
        optChar, low, high);
    PrintError(errorStr);
    AreParamsValid = false;
  }
}

/* Exit if a file already exists */
void AssertFileDNE(const char *filename) {
  char errorStr[64];

  if (access(filename, F_OK) != -1) {
    sprintf(errorStr,
        "ERROR: File '%s' already exists\n", filename);
    PrintError(errorStr);
    AreParamsValid = false;
  }
}


/* Allow the user to change simulation parameters via the command line

   @param argc The number of command line arguments to parse
   @param argv The array of command line arguments to parse
   */
void GetUserOptions(const int argc, char **argv) {
  char c; /* Loop control variable */

  /* Loop over argv, setting parameter values given */
  while ((c = getopt(argc, argv, GETOPT_STRING)) != -1) {
    switch(c) {
      case ROWS_CHAR:
        Rows = atoi(optarg);
        AssertPositiveInteger(Rows, ROWS_CHAR);
        break;
      case COLS_CHAR:
        Cols = atoi(optarg);
        AssertPositiveInteger(Cols, COLS_CHAR);
        break;
      case BURN_PROB_CHAR:
        BurnProb = atoi(optarg);
        AssertPositiveInteger(BurnProb, BURN_PROB_CHAR);
        AssertBetweenInclusive(BurnProb, 0, 100, BURN_PROB_CHAR);
        break;
      case MAX_BURN_STEPS_CHAR:
        MaxBurnSteps = atoi(optarg);
        AssertPositiveInteger(MaxBurnSteps, MAX_BURN_STEPS_CHAR);
        break;
      case STEPS_CHAR:
        Steps = atoi(optarg);
        AssertPositiveInteger(Steps, STEPS_CHAR);
        break;
      case RAND_SEED_CHAR:
        RandSeed = atoi(optarg);
        AssertPositiveInteger(RandSeed, RAND_SEED_CHAR);
        break;
      case OUTPUT_FILENAME_CHAR:
        IsOutputtingEachStep = true;
        OutputFilename = optarg;
        break;
      case '?':
      default:
        PrintError("ERROR: illegal option\n");
        PrintUsageAndExit();
    }
  }

  if (IsOutputtingEachStep) { /* See if we need to make sure the output file
                                 does not exist (DNE) */
    if (MyRank == BOSS) { /* There would be a race condition if one process
                             checks if the file exists, sees that it does not
                             exist (DNE), and creates the file, then another
                             process checks if the file exists and sees that it 
                             does exist, even though it DNE initially */
      AssertFileDNE(OutputFilename);
    }
  }

  if (Procs > Rows) {
    PrintError("ERROR: The number of processes cannot be greater than the number of rows\n");
    AreParamsValid = false;
  }
}

/* Calculate the number of rows for which each process is responsible */
void DistributeRows() {
  const int evenSplit = Rows / Procs; /* If the number of rows is divided
                                         evenly with no remainder, this is
                                         the result */
  const int remainder = Rows % Procs; /* The remainder of that division is
                                         this result */

  /* All processes have at least an even split of the forest */
  Rows = evenSplit;

  /* The remainder is split up evenly among the first few processes; each
     process gets one extra row */
  if (MyRank < remainder) {
    Rows++;
  }
}

/* Allocate dynamic memory */
void AllocateMemory() {
  /* Two extra rows and columns are needed for the boundary */
  Trees    = (int*)malloc((Rows + 2) * (Cols + 2) * sizeof(int));
  NewTrees = (int*)malloc(Rows * Cols * sizeof(int));

  /* The send and receive buffers are single rows with two boundary columns
     each */
  SendBuf = (int*)malloc((Cols + 2) * sizeof(int));
  RecvBuf = (int*)malloc((Cols + 2) * sizeof(int));
}

/* @param row The tree's row
   @param col The tree's column
   @return The number of steps the tree at the row and column has been burning
   in the current time step
   */
int GetTree(const int row, const int col) {
  /* Have to index the tree with boundaries included */
  return Trees[TREE_MAP(row, col, Cols)];
}

/* @param row The tree's row
   @param col The tree's column
   @return The number of steps the tree at the row and column has been burning
   in the next time step
   */
int GetNewTree(const int row, const int col) {
  /* Have to index the tree with boundaries included */
  return NewTrees[NEW_TREE_MAP(row, col, Cols)];
}

/* Set number of time steps burned for a tree in the current time step

   @param row The tree's row
   @param col The tree's column
   @param burnSteps The number of time steps to set the tree as having burned
   so far in the current time step */
void SetTree(const int row, const int col, const int burnSteps) {
  Trees[TREE_MAP(row, col, Cols)] = burnSteps;
}

/* Set number of time steps burned for a tree in the next time step

   @param row The tree's row
   @param col The tree's column
   @param burnSteps The number of time steps to set the tree as having burned
   so far in the current time step */
void SetNewTree(const int row, const int col, const int burnSteps) {
  NewTrees[NEW_TREE_MAP(row, col, Cols)] = burnSteps;
}

/* Light a random tree on fire, set all other trees to be not burning */
void InitData() {
  int row;
  int col;
  int rank;

  /* Set all trees as having burned for 0 time steps */
#pragma omp parallel for private(col)
  for (row = 1; row < Rows + 1; row++) {
    for (col = 1; col < Cols + 1; col++) {
      SetTree(row, col, 0);
      SetNewTree(row, col, 0);
    }
  }

  /* Set the boundaries as burnt out */
#pragma omp parallel for private(col)
  for (row = 0; row < Rows + 2; row++) {
    /* Left */
    SetTree(row, 0, MaxBurnSteps);

    /* Top/Bottom */
    if ((MyRank == 0 && row == 0) ||
        (MyRank == Procs - 1 && row == Rows + 1)) {
      for (col = 1; col < Cols + 1; col++) {
        SetTree(row, col, MaxBurnSteps);
      }
    }

    /* Right */
    SetTree(row, Cols + 1, MaxBurnSteps);
  }

  /* The boss chooses a random process to light a random tree on fire */
  if (MyRank == BOSS) {
    rank = Procs * random() / RAND_MAX;
  }

  /* The boss broadcasts the rank of the random process */
  MPI_Bcast(&rank, 1, MPI_INT, BOSS, MPI_COMM_WORLD);

  /* The chosen process sets a random tree as having burned for 1 time step */
  if (rank == MyRank) {
    /* Set a random tree as having burned for 1 time step */
    row = 1 + (Rows + 1) * random() / RAND_MAX;
    col = 1 + (Cols + 1) * random() / RAND_MAX;
    SetTree(row, col, 1);
    SetNewTree(row, col, 1);
    BurnedTrees++; /* Update the tally of trees that have burned */
  }
}

/* Output tree data for the current time step */
void OutputData() {
  int row;
  int col;
  char buf[64];

  /* Write the header for the current time step */
  sprintf(buf, "Time step %d\n", CurStep);
  fprintf(OutputFile, "%s", buf);

  for (row = 1; row < Rows + 1; row++) {
    for (col = 1; col < Cols + 1; col++) {
      sprintf(buf, "%*d ", MaxBurnStepsDigits, GetTree(row, col));
      fprintf(OutputFile, "%s", buf);
    }
    fprintf(OutputFile, "\n");
  }

  fprintf(OutputFile, "\n");
}

/* @return Whether the tree in the given row and column has burnt out */
bool IsBurntOut(const int row, const int col) {
  return GetTree(row, col) >= MaxBurnSteps;
}

/* @return Whether the tree in the given row and column is on fire */
bool IsOnFire(const int row, const int col) {
  return GetTree(row, col) > 0 && !IsBurntOut(row, col);
}

/* For trees already burning, increment the number of time steps they have
   burned */
void ContinueBurning() {
  int row;
  int col;

#pragma omp parallel for private(col)
  for (row = 1; row < Rows + 1; row++) {
    for (col = 1; col < Cols + 1; col++) {
      if (IsOnFire(row, col)) {
        SetNewTree(row, col, GetTree(row, col) + 1);
      }
    }
  }
}

/* Send a row to a process then receive a row from the same process,
   using the current time step as the tag of the messages

   @param rank The rank of the process to send/receive to/from
   @param rank The index of the row to send
   @param rank The index of the row to receive
   */
void SendRecv(const int rank, const int sendRow, const int recvRow) {
  int col;

  /* Before sending, fill the send buffer with the row to send */
  for (col = 0; col < Cols + 2; col++) {
    SendBuf[col] = GetTree(sendRow, col);
  }

  MPI_Sendrecv(SendBuf, /* send buffer */
      Cols + 2, /* number of elements to send */
      MPI_INT, /* type of element to send */
      rank, /* rank of process to send to */
      CurStep, /* send message tag */
      RecvBuf, /* receive buffer */
      Cols + 2, /* number of elements to receive */
      MPI_INT, /* type of element to receive */
      rank, /* rank of process to receive from */
      CurStep, /* receive messsage tag */
      MPI_COMM_WORLD, /* communicator */
      MPI_STATUS_IGNORE); /* status */

  /* After receiving, fill the row to receive with the receive buffer */
  for (col = 0; col < Cols + 2; col++) {
    SetTree(recvRow, col, RecvBuf[col]);
  }
}

/* Send/receive boundary rows to/from neighbor processes */
void CommunicateBoundaries() {
  /* Communicate with the top neighbor */
  if (MyRank > 0) { /* The top process has no top neighbor */
    SendRecv(MyRank - 1, /* process to communicate with */
        1, /* index of row to send */
        0); /* index row to receive */
  }

  /* Communicate with the bottom neighbor */
  if (MyRank < Procs - 1) { /* The bottom process has no bottom neighbor */
    SendRecv(MyRank + 1, /* process to communicate with */
        Rows, /* index of row to send */
        Rows + 1); /* index of row to receive */
  }
}

/* Find trees that are not on fire yet and try to catch them on fire from
   burning neighbor trees */
void BurnNew() {
  int row;
  int col;

#pragma omp parallel for private(col)
  for (row = 1; row < Rows + 1; row++) {
    for (col = 1; col < Cols + 1; col++) {
      if (!IsOnFire(row, col) && !IsBurntOut(row, col)) {
        /* Check neighbors */
        if ((IsOnFire(row-1, col) || /* Top */
              IsOnFire(row, col-1) || /* Left */
              IsOnFire(row+1, col) || /* Bottom */
              IsOnFire(row, col+1)) && /* Right */
            ((100 * random() / RAND_MAX) < BurnProb)) { /* Random chance */
          SetNewTree(row, col, 1); /* Catch the tree on fire */
#pragma omp atomic
          BurnedTrees++; /* Update the tally of trees that have burned */
        }
      }
    }
  }
}

/* Copy new tree data into old tree data */
void AdvanceTime() {
  int row;
  int col;

#pragma omp parallel for private(col)
  for (row = 1; row < Rows + 1; row++) {
    for (col = 1; col < Cols + 1; col++) {
      SetTree(row, col, GetNewTree(row, col));
    }
  }
}

/* Free allocated memory */
void FreeMemory() {
  free(RecvBuf);
  free(SendBuf);
  free(NewTrees);
  free(Trees);
}

/* @param argc The number of command line arguments
   @param argv String of command line arguments
   */
int main(int argc, char **argv) {
  int trees; /* Number of trees */
  int i;

  /* Initialize MPI */
  MPI_Init(&argc, &argv);

  /* Get the MPI rank of the process executing this code */
  MPI_Comm_rank(MPI_COMM_WORLD, &MyRank);

  /* Get the total number of MPI processes */
  MPI_Comm_size(MPI_COMM_WORLD, &Procs);

  /* The boss process sets the program executable name */
  if (MyRank == BOSS) {
    /* Set the program executable name */
    strcpy(ExeName, argv[0]);
  }

  /* Allow the user to change simulation parameters via the command line */
  GetUserOptions(argc, argv);

  /* Since the boss might have done some validity checks that the non-bosses
     didn't, the boss has the final say on whether parameters were valid, so
     the boss broadcasts this information to the others */
  MPI_Bcast(&AreParamsValid, /* Send/recv buffer */
      1, /* Number of elements */
      MPI_INT, /* Type of elements */
      BOSS, /* Sender */
      MPI_COMM_WORLD); /* Communicator */
  if (!AreParamsValid) {
    /* Model parameters are not valid; exit early */
    PrintUsageAndExit();
  }

  /* Calculate total number of trees */
  trees = Rows * Cols;

  if (IsOutputtingEachStep) {
    /* Open the output file */
    OutputFile = fopen(OutputFilename, "w");
  }

  /* Calculate the number of rows for which each process is responsible */
  DistributeRows();

  /* Seed the random number generator */
  srandom(RandSeed);

  /* Allocate dynamic memory for the 1D tree arrays */
  AllocateMemory();

  /* Light a random tree on fire, set all other trees to be not burning */
  InitData();

  /* Start the simulation looping for the specified number of time steps */
  for (CurStep = 0; CurStep < Steps; CurStep++) {
    if (IsOutputtingEachStep) {
      /* Output tree data for the current time step */
      OutputData();
    }

    /* For trees already burning, increment the number of time steps they have
       burned */
    ContinueBurning();

    /* Send/receive boundary rows to/from neighbor processes */
    CommunicateBoundaries();

    /* Find trees that are not on fire yet and try to catch them on fire from
       burning neighbor trees */
    BurnNew();

    /* Copy new tree data into old tree data */
    AdvanceTime();
  }

  /* Print the total percentage of trees burned */
  MPI_Reduce(&BurnedTrees, /* Send buffer */
      &i, /* Receive buffer */
      1, /* Number of elements */
      MPI_INT, /* Type of elements */
      MPI_SUM, /* Type of operation */
      BOSS, /* Receiver */
      MPI_COMM_WORLD); /* Communicator */

  if (MyRank == BOSS) { /* The boss propares to print the percentage of trees
                           burned */
    BurnedTrees = i;
    printf("%.2f%% of the trees were burned\n", (1.0 * BurnedTrees) / trees);
  }

  if (IsOutputtingEachStep) {
    /* Close the output file */
    fclose(OutputFile);
  }

  /* Free allocated memory */
  FreeMemory();

  /* Clean up MPI */
  MPI_Finalize();

  return 0;
}

